////////////////////////////////////////////////////////////////// COPYRIGHT //
//
//  XmdvTool Version 4.0 - A program for Multivariate Data Display
//
//  Copyright 1994-1999, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//      Matthew O. Ward, Allen R. Martin, Ying-Huey Fua
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/~xmdv
//
//  Send bug reports, bug fixes, enhancements, requests, or comments to the
//  XmdvTool Development Team: <matt@cs.wpi.edu>.
//
////////////////////////////////////////////////////////////////// COPYRIGHT //

#include "cluster_parcoord.H"
#include "ogl.H"
#include "cluster.H"
#include "XmdvTool.H"
#include "colormap.H"
#include "parcoord.H"
#include "magnify.H"
#include "brush_oper.H"

float cluster_radius = .5;
float fade_factor=0; // [-1,+1]
float extent_scaling_fac=1.0;

colormap CMAP(true);

bool SHOW_BANDS = false;

static bool SHOW_MEANS = true;
static bool SHOW_BRUSH_BKGRD = true;
static bool COLORED_BANDS = true;

//bool EXTENT_BRUSH = true;

enum COLOR_STYLE { STYLE_NONE=0x01, STYLE_HIGHLIGHT=0x02, STYLE_COLORMAP=0x04, STYLE_THICK=0x08, STYLE_DIM=0x10 };
int normal_style = STYLE_COLORMAP;
int highlight_style = STYLE_HIGHLIGHT | STYLE_THICK;

#define XMAP(I) (((I)+.5) * dataset->one_over_dims)
#define YMAP(X,Y) (1.0-((Y) - dataset->dim_min[X]) / (dataset->dim_max[X] - dataset->dim_min[X]))
#define YINVMAP(X,Y) (((1.0-(Y)) * (dataset->dim_max[X] - dataset->dim_min[X])) + dataset->dim_min[X])
#define YGAP(X,Y) ((Y)/(dataset->dim_max[X] - dataset->dim_min[X]))

void RegisterHierParcoordCanvas(grCanvas *canvas)
{
    canvas->display_proc = do_cluster_parcoord;
    canvas->reshape_proc = grCanvas::grResizeMainCanvas;
    canvas->motion_proc = InputHierParcoordMouseMotion;
    canvas->buttonpress_proc = InputHierParcoordButtonPress;
    canvas->buttonrelease_proc = InputHierParcoordButtonRelease;
//    canvas->keypress_proc = appCanvasInputKeyPress;
    canvas->keyrelease_proc = InputHierParcoordKeyRelease;
}

static void hierDrawSingleNodeMean( grCanvas *canvas, Cluster &c, int style)
{
    ARR<Vec2> poly(dataset->dims);
    for (int d=0; d<dataset->dims; d++)
    {
        poly[d].X = XMAP(d);
        poly[d].Y = YMAP(d,c.mean[d]);
    }

    if (BITTEST(style,STYLE_NONE)) return;

    float opacity = 1.0;

//  if (BITTEST(style,STYLE_DIM)) opacity *= 0.75;

    if (BITTEST(style,STYLE_HIGHLIGHT)) {

        // Want to fade out the brushed lines
        if (fade_factor<0) opacity = 1+fade_factor;

        canvas->SetForeground(interface_color[SBB], opacity);
    } else
    if (BITTEST(style,STYLE_COLORMAP)) {

        // Want to fade out the unbrushed lines
        if (fade_factor>0) opacity=1-fade_factor;

        // Check if mean is covered by data brush.
        if ( c.mark == okcdata::COLORED )
            canvas->SetForeground(interface_color[HIGHLIGHT], opacity);
        else {
            int color_idx = int(c.color*(CMAPSIZE-1));
            canvas->SetForeground( CMAP[color_idx], opacity);
        }
    }

    if (BITTEST(style,STYLE_THICK)) {
        glLineWidth(1.5);
    } else {
        glLineWidth(1.0);
    }

    canvas->DrawLines(poly);

    glLineWidth(1.0);
}

static void hierDrawSingleNodeBand( grCanvas *canvas, Cluster &c, float density)
{
    bool is_brushed = BITTEST(c.flags,IS_BRUSHED);
    if (MAGNIFY && !is_brushed) return;

    if (COLORED_BANDS) {

        int color_idx = (int)(c.color*(CMAPSIZE-1));

        if (is_brushed)  {
            DB(fade_factor);
            if (fade_factor>0)
                canvas->SetForeground( CMAP[ color_idx ], 1.0);
            else
                canvas->SetForeground( CMAP[ color_idx ], 1+fade_factor);
        }
        else {
            if (fade_factor>0)
                canvas->SetForeground( CMAP[ color_idx ], 1-fade_factor);
            else
                canvas->SetForeground( CMAP[ color_idx ], 1.0);
        }
    }
    else
        canvas->SetForeground(interface_color[BRUSH1+2]);


    // Draw extent bands
    //
    ARR<Vec2> band(4);

    // Draw the lower band
    //
    int d;
    for (d=0; d<dataset->dims; d++)
    {
        double radius = c.sqrt_radius * (dataset->all_dim_max[d]-dataset->all_dim_min[d]);

        Vec2 lo, hi;

        lo.X = XMAP(d);
        lo.Y = YMAP(d,c.mean[d]- radius);

        hi.X = XMAP(d);
        hi.Y = YMAP(d,c.mean[d]);

        lo.Y = (1-extent_scaling_fac)*hi.Y + (extent_scaling_fac*lo.Y);

        if (d==0) {
            band[0] = lo;
            band[3] = hi;

        } else {

            band[1] = lo;
            band[2] = hi;
            canvas->DrawTransparentBand2(band, 0.0, density, 0);

            band[0] = band[1];
            band[3] = band[2];
        }
    }

    // Draw the upper band
    //
    for (d=0; d<dataset->dims; d++)
    {
        double radius = c.sqrt_radius * (dataset->all_dim_max[d]-dataset->all_dim_min[d]);

        Vec2 lo, hi;

        lo.X = XMAP(d);
        lo.Y = YMAP(d,c.mean[d]);

        hi.X = XMAP(d);
        hi.Y = YMAP(d,c.mean[d]+ radius);

        // Calculate how much to scale, if any
        hi.Y = (1-extent_scaling_fac)*lo.Y + (extent_scaling_fac*hi.Y);

        if (d==0) {
            band[0] = lo;
            band[3] = hi;
        } else {
            band[1] = lo;
            band[2] = hi;

            canvas->DrawTransparentBand2(band, density, 0.0, 0);

            band[0] = band[1];
            band[3] = band[2];
        }
    }
}

#define SCAN_FLAGS(F,ON,OFF) ((((F)&ON)==ON) && (((0xffff^(F))&OFF)==OFF))

static void hierDrawColoredBands(grCanvas *canvas, int on_flags=0, int off_flags=0)
{
    for (int i=0;i<btree->nodes.Len();i++)
    {
        Cluster &c = btree->nodes[i];
        if (!SCAN_FLAGS(c.flags, on_flags, off_flags)) continue;

        hierDrawSingleNodeBand(canvas, c, 0.5);
    }
}

static void hierDrawColoredMeans(grCanvas *canvas, int style, int on_flags=0, int off_flags=0 )
{
    for (int i=0;i<btree->nodes.Len();i++)
    {
        Cluster &c = btree->nodes[i];
        if (!SCAN_FLAGS(c.flags, on_flags, off_flags)) continue;

        hierDrawSingleNodeMean(canvas, c, style);
    }
}

void do_cluster_parcoord( struct grCanvas *canvas )
{
    if (!dataset) return;

    canvas->SetTopLeftOrigin();
    grCanvas::grResizeMainCanvas(canvas);

    canvas->SetBackground(interface_color[BG]);
    canvas->Clear();

    hierInitClusterTree();

    if (!btree) return;

    //if (SHOW_BRUSH_BKGRD)
    DrawParcoordBrushes(canvas);

    if (SHOW_BANDS) {
        if (!MAGNIFY) hierDrawColoredBands(canvas, IS_CONTOUR0);
        hierDrawColoredBands(canvas, IS_CONTOUR1);
    }

    if (SHOW_MEANS) {
        if (!MAGNIFY) hierDrawColoredMeans(canvas, normal_style, IS_CONTOUR0, IS_BRUSHED );
        hierDrawColoredMeans(canvas, highlight_style, IS_CONTOUR1 | IS_BRUSHED );
    }

    // Draw average operations
    for (int op=0; op<num_oper; op++)
    {
        // Check if this operation has OPER_AVERAGE set */
        if (!(brush_ops[op].oper & AVERAGE)) continue;

        ARR<float> average(dataset->dims);
        if (BrushOperCompAverage(op, average)==0) continue;

        glLineWidth(3.0);
        canvas->SetForeground(interface_color[TEXT]);
        DrawParcoordAverage(canvas,average);
    }

    glLineWidth(1.);

    DrawParcoordAxes(canvas);
}

void InputHierParcoordMouseMotion(grCanvas *canvas, int x, int y)
{
    // Hack to disable multiselect in HPC
    if (canvas->shiftkey && canvas->button[0] == true) return;

    InputParcoordMouseMotion(canvas,x,y);
}

void InputHierParcoordButtonPress(grCanvas *canvas, int button, int x, int y)
{
    // Hack to disable multiselect in HPC
    if (canvas->shiftkey && button == 0) return;

    InputParcoordButtonPress(canvas,button,x,y);
}

void InputHierParcoordButtonRelease(grCanvas *canvas, int button, int x, int y)
{
    // Hack to disable multiselect in HPC
    if (canvas->shiftkey && button == 0) return;

    InputParcoordButtonRelease(canvas,button,x,y);

    if (canvas->shiftkey) return;

    if (!btree) return;

//    EXTENT_BRUSH = true;
//    hierClearNodeFlags(IS_BRUSHED);
//    hierModifyExtentsByBrushParameters();
//    hierMarkBrushedClustersByExtents(&btree->nodes[0]);
    canvas->Refresh();
}

#include "X11/keysym.h"

void InputHierParcoordKeyRelease(grCanvas *canvas, int key)
{
    if (key == XK_Shift_L || key == XK_Shift_R) return;

    appCanvasInputKeyRelease(canvas,key);

    if (key == XK_Shift_L || key == XK_Shift_R) {
        // Exit multiselect mode
        //
//        EXTENT_BRUSH = true;
//        hierClearNodeFlags(IS_BRUSHED);
//        hierModifyExtentsByBrushParameters();
//        hierMarkBrushedClustersByExtents(&btree->nodes[0]);
        canvas->Refresh();
    }
}
