////////////////////////////////////////////////////////////////// COPYRIGHT //
//
//  XmdvTool Version 4.0 - A program for Multivariate Data Display
//
//  Copyright 1994-1999, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//      Matthew O. Ward, Allen R. Martin, Ying-Huey Fua
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/~xmdv
//
//  Send bug reports, bug fixes, enhancements, requests, or comments to the
//  XmdvTool Development Team: <matt@cs.wpi.edu>.
//
////////////////////////////////////////////////////////////////// COPYRIGHT //

#include "okcdata.H"
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "macros.H"

static const char *read_till(FILE *fp, const char *delimiters="\n", char *which_delimiter=0)
{
    static char str[500];
    int k=0;
    while (1) {
        char c=fgetc(fp);
        char *ptr = strchr(delimiters, c);
        if (ptr) {
            str[k] = '\0';
            if (which_delimiter) *which_delimiter = *ptr;
            break;
        }
        str[k++] = c;
    }
    return str;
}

// Read in data file: Format is:
// number_of_dimensions number_of_samples (int int)
// dimension_name for each dimension (string)
// minimum maximum cardinality for each dimension (float float int)
// data_samples (floats)
//
void okcdata::Read(const char *fname)
{
    FILE *fp = fopen(fname, "r");
    if(fp == NULL)
    {
        fprintf(stderr,"Data file %s doesn't exist\n", fname);
        return;
    }

    datafile = strdup(fname);

    // Get number of dimensions, and number of records
    //
    if (fscanf(fp, "%d %d\n", &all_dims, &data_size) <= 0)  {
        printf("Error reading first line of file %s!\n", fname);
        printf("Try loading another file!\n");
        all_dims = 0;
        data_size = 0;
    }

    if (data_size < 0)  {
        printf("Number of data items is negative (%d)!\n", data_size);
        printf("Try loading another file!\n");
        all_dims = 0;
        data_size = 0;
    }

    // Read in names of dimensions, and if necessary, nominal
    // encodings of values.
    //
    all_names.SetLen(all_dims);
    all_nominal_fields.SetLen(all_dims);

    int i;
    for (i=0; i < all_dims; i++)
    {
        char endchar;
        const char *dname = read_till(fp,"(\n",&endchar);
        all_names[i] = new char[strlen(dname)+4];
        strcpy(all_names[i],dname);

        NominalField *ndata = 0;
        if (endchar == '(') {
            ndata = new NominalField;
            ndata->range_min = atoi(read_till(fp,"),\n",&endchar));
            ndata->range_max = atoi(read_till(fp,"),\n",&endchar));
            ndata->values.SetLen(ndata->range_max-ndata->range_min+1);

            int n;
            for (n=0;n<ndata->values.Len();n++) {
                ndata->values[n] = 0;
            }

            if (endchar != ')') {
                for (n=ndata->range_min; n<=ndata->range_max; n++) {
                    const char *str = read_till(fp,",)\n",&endchar);
                    ndata->values[n-ndata->range_min] = strdup(str);
                }
            }

            read_till(fp,"\n");
        }

        all_nominal_fields[i] = ndata;
    }

    all_dim_min.SetLen(all_dims);
    all_dim_max.SetLen(all_dims);
    all_cardinality.SetLen(all_dims);
    on.SetLen(all_dims);

    for (i=0;i < all_dims;i++)
    {
        if (fscanf(fp, "%f %f %d", &all_dim_min[i], &all_dim_max[i], &all_cardinality[i]) !=3) {
            printf("failed in reading dimension ranges.\n");
            exit(-1);
        }

        // Need to handle case when min = max (separate them a bit)
        //
        if (all_dim_min[i] == all_dim_max[i]) {
            if (all_dim_min[i] == 0.0)  {
                all_dim_min[i] = -1.0;
                all_dim_max[i] = 1.0;
            } else {
                if (all_dim_min[i] < 0.0) {
                    all_dim_min[i] = 2. * all_dim_min[i];
                    all_dim_max[i] = 0.0;
                } else {
                    all_dim_max[i] = 2. * all_dim_min[i];
                    all_dim_min[i] = 0.0;
                }
            }
        } else {

            // Add a bit of space to the bounds to avoid boundary problems
            //
            const float SPACE=.05; // extra space for bounds to avoid landing on edges

            float dim_diff = (all_dim_max[i] - all_dim_min[i]) * SPACE;
            all_dim_max[i] += dim_diff;
            all_dim_min[i] -= dim_diff;
        }

        on[i] = true;
    }

    data_buf.SetLen(all_dims*data_size);

    int k=0;
    for (i=0;i < data_size;i++) {
        for (int j = 0;j < all_dims;j++,k++) {
            if (fscanf(fp, "%f", &data_buf[k]) != 1) {
                printf("failed in reading data, line %d, item %d.\n", i, j);
            }
        }
    }

    fclose(fp);

    mark.SetLen(data_size);
    color.SetLen(data_size);

    SetFieldOnOff(-1,true);
}

void okcdata::GetData(ARR<float> &buf, int line)
{
    int j=0;
    int base = all_dims*line;
    for (int i=0;i<all_dims;i++)
        if (on[i])
            buf[j++] = data_buf[base + i];
}

#include "brush.H"

void okcdata::SetFieldOnOff(int which, bool state)
{
    int br;
    for (br=0; br<MAXBRUSH; br++)
    {
        int j = 0;
        for (int i=0;i<MIN(brushes[br].pos.Len(),on.Len());i++)
            if (on[i]) {
                all_brushes[br].pos[i] = brushes[br].pos[j];
                all_brushes[br].size[i] = brushes[br].size[j];
                all_brushes[br].bound_offset[i] = brushes[br].bound_offset[j];
                j++;
            }
    }

    if (which>=0) on[which] = state;

    dims = all_dims;
    names.SetLen(all_dims);
    dim_min.SetLen(all_dims);
    dim_max.SetLen(all_dims);
    nominal_fields.SetLen(all_dims);
    org_dim_min.SetLen(all_dims);
    org_dim_max.SetLen(all_dims);

    cardinality.SetLen(all_dims);

    dims = 0;
    for (int i=0;i<all_dims;i++)
        if (on[i]) {
            org_dim_min[dims] = dim_min[dims] = all_dim_min[i];
            org_dim_max[dims] = dim_max[dims] = all_dim_max[i];
            cardinality[dims] = all_cardinality[i];
            names[dims] = all_names[i];
            nominal_fields[dims] = all_nominal_fields[i];
            dims++;
        }


    one_over_dims = 1.0/dims;

    for (br=0; br<MAXBRUSH; br++)
    {
        int j = 0;
        for (int i=0;i<MIN(brushes[br].pos.Len(),on.Len());i++)
            if (on[i]) {
                brushes[br].pos[j] = all_brushes[br].pos[i];
                brushes[br].size[j] = all_brushes[br].size[i];
                brushes[br].bound_offset[j] = all_brushes[br].bound_offset[i];
                j++;
            }
    }
}

const char *NominalField::MapValueToString(int what)
{
    if (what<range_min || what>range_max || values[what-range_min]==0) {
        static char number_string[10];
        sprintf(number_string,"%d",what);
        return number_string;
    }
    return values[what-range_min];
}
